#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "gtk/gtk.h"
#include "glib.h"

enum game_state {
	GAME_STATE_START,
	GAME_STATE_CHOSEN,
	GAME_STATE_REVEALED,
	GAME_STATE_FINISH
};

struct game {
	enum game_state state;
	unsigned int num_trials;
	unsigned int num_stayed;
	unsigned int won_stayed;
	unsigned int num_switched;
	unsigned int won_switched;
	unsigned int num_doors;
	unsigned int *doors;
	unsigned int prize_door;
	unsigned int chosen_door;
	unsigned int unopened_door;
	unsigned int switched;
};

void game_init(struct game *game, unsigned int num_doors)
{
	game->num_trials = 0;
	game->num_stayed = 0;
	game->won_stayed = 0;
	game->num_switched = 0;
	game->won_switched = 0;
	game->num_doors = num_doors;
	game->doors = malloc(sizeof(*game->doors) * game->num_doors);
	if (game->doors == NULL)
		exit(1);

	srand(time(NULL));
}

void game_free(struct game *game)
{
	free(game->doors);
}

void game_close_doors(struct game *game)
{
	unsigned int i;

	for (i = 0; i < game->num_doors; i++)
		game->doors[i] = 0;
}

void game_place_prize(struct game *game)
{
	game->prize_door = rand() % game->num_doors;
}

void game_start_round(struct game *game)
{
	game->state = GAME_STATE_START;
	game_close_doors(game);
	game_place_prize(game);
	game->switched = 0;
}

void game_choose_door(struct game *game, unsigned int n)
{
	assert(game->state == GAME_STATE_START);

	game->state = GAME_STATE_CHOSEN;
	game->chosen_door = n;
}

void game_open_door(struct game *game, unsigned int n)
{
	game->doors[n] = 1;
}

unsigned int game_door_is_open(struct game *game, unsigned int n)
{
	return !!game->doors[n];
}

void game_reveal_doors(struct game *game)
{
	unsigned int i;

	assert(game->state == GAME_STATE_CHOSEN);

	game->state = GAME_STATE_REVEALED;
	if (game->chosen_door == game->prize_door) {
		game->unopened_door = rand() % (game->num_doors - 1);
		if (game->unopened_door == game->chosen_door)
			game->unopened_door++;
	} else {
		game->unopened_door = game->prize_door;
	}

	for (i = 0; i < game->num_doors; i++) {
		if (i != game->chosen_door && i != game->unopened_door)
			game_open_door(game, i);
	}
}

void game_switch(struct game *game)
{
	unsigned int tmp;

	assert(game->state == GAME_STATE_REVEALED);

	tmp = game->chosen_door;
	game->chosen_door = game->unopened_door;
	game->unopened_door = tmp;

	game->switched = !game->switched;
}

unsigned int game_is_won(struct game *game)
{
	return game->chosen_door == game->prize_door;
}

void game_finish_round(struct game *game)
{
	game_open_door(game, game->chosen_door);
	game_open_door(game, game->unopened_door);

	assert(game->state == GAME_STATE_REVEALED);

	game->state = GAME_STATE_FINISH;
	game->num_trials++;
	if (game->switched) {
		game->num_switched++;
		if (game_is_won(game))
			game->won_switched++;
	} else {
		game->num_stayed++;
		if (game_is_won(game))
			game->won_stayed++;
	}
}

void game_print_report(struct game *game)
{
	printf("Round %u: Stay: %u / %u = %.3f Switch: %u / %u = %.3f\n",
		game->num_trials + 1,
		game->won_stayed, game->num_stayed,
		(game->num_stayed == 0) ? 0.0 : (double) game->won_stayed / game->num_stayed,
		game->won_switched, game->num_switched,
		(game->num_switched == 0) ? 0.0 : (double) game->won_switched / game->num_switched);
}

/* *** */

#define NUM_DOORS 3
#define DOOR_WIDGET_WIDTH 80
#define DOOR_WIDGET_HEIGHT 100

enum door_state {
	DOOR_CLOSED,
	DOOR_SELECTED,
	DOOR_OPEN,
	DOOR_OPEN_PRIZE
};

static const char *DOOR_IMAGE_FILENAMES[] = {
	"door-closed.png",
	"door-selected.png",
	"door-open.png",
	"door-prize.png"
};

static GdkPixbuf *door_pixbufs[] = {
	NULL,
	NULL,
	NULL,
	NULL
};

static struct game game;
static GtkButton *doors[NUM_DOORS];
static GtkWidget *announce_label, *summary_label;

static void update_doors(GtkButton *doors[], struct game *game)
{
	GtkButton *door;
	GtkImage *image;
	unsigned int i;

	for (i = 0; i < NUM_DOORS; i++) {
		door = doors[i];
		g_object_get(door, "image", &image, NULL);
		if (game_door_is_open(game, i)) {
			if (i == game->prize_door)
				gtk_image_set_from_pixbuf(image, door_pixbufs[DOOR_OPEN_PRIZE]);
			else
				gtk_image_set_from_pixbuf(image, door_pixbufs[DOOR_OPEN]);
		} else {
			if (i == game->chosen_door && game->state != GAME_STATE_START)
				gtk_image_set_from_pixbuf(image, door_pixbufs[DOOR_SELECTED]);
			else
				gtk_image_set_from_pixbuf(image, door_pixbufs[DOOR_CLOSED]);
		}
		g_object_unref(image);
	}
}

static gboolean delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gtk_main_quit();
	return FALSE;
}

static void summary_label_update()
{
	char *s;

	s = g_markup_printf_escaped(
		"Round %u\nStaying won %u / %u = %.2f%%\nSwitching won %u / %u = %.2f%%",
		game.num_trials + 1, game.won_stayed, game.num_stayed,
		(game.num_stayed == 0) ? 0.0 : (double) 100.0 * game.won_stayed / game.num_stayed,
		game.won_switched, game.num_switched,
		(game.num_switched == 0) ? 0.0 : (double) 100.0 * game.won_switched / game.num_switched);
	gtk_label_set_text(GTK_LABEL(summary_label), s);
	g_free(s);
}

static void door_clicked(GtkButton *button, gpointer pn)
{
	unsigned int n;

	n = GPOINTER_TO_UINT(pn);
	if (game.state == GAME_STATE_START) {
		game_choose_door(&game, n);
		game_reveal_doors(&game);
	} else if (game.state == GAME_STATE_CHOSEN) {
	} else if (game.state == GAME_STATE_REVEALED) {
		if (n == game.chosen_door) {
			game_finish_round(&game);
		} else if (n == game.unopened_door) {
			game_switch(&game);
			game_finish_round(&game);
		}
		summary_label_update();
	} else if (game.state == GAME_STATE_FINISH) {
		game_start_round(&game);
	}
	update_doors(doors, &game);
}

static GtkWidget *make_door_widget(unsigned int n)
{
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *image;
	GtkWidget *label;
	char *s;

	button = gtk_button_new();
	gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
	vbox = gtk_vbox_new(FALSE, 0);
	image = gtk_image_new_from_pixbuf(door_pixbufs[DOOR_CLOSED]);
	gtk_box_pack_start(GTK_BOX(vbox), image, TRUE, FALSE, 0);
	gtk_widget_show(image);
	g_object_set(G_OBJECT(button), "image", image);
	s = g_markup_printf_escaped("%u", n);
	label = gtk_label_new(s);
	g_free(s);
	gtk_box_pack_end(GTK_BOX(vbox), label, TRUE, FALSE, 0);
	gtk_widget_show(label);
	gtk_widget_show(vbox);
	/*
	gtk_widget_set_size_request(button,
		DOOR_WIDGET_WIDTH, DOOR_WIDGET_HEIGHT);
	*/
	gtk_container_add(GTK_CONTAINER(button), vbox);
	gtk_widget_show(button);

	return button;
}

int main(int argc, char *argv[])
{
	GtkWidget *window;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *box;
	GtkWidget *door;
	GdkColor bg_color = { 0, 0.8 * 65535, 0.9 * 65535, 1.0 * 65535 };
	unsigned int i;

	game_init(&game, NUM_DOORS);
	game_start_round(&game);

	gtk_init(&argc, &argv);

	for (i = 0; i < sizeof(door_pixbufs) / sizeof(*door_pixbufs); i++)
		door_pixbufs[i] = gdk_pixbuf_new_from_file(DOOR_IMAGE_FILENAMES[i], NULL);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(window), "delete_event",
		G_CALLBACK(delete_event), NULL);
	gtk_widget_modify_bg(window, GTK_STATE_NORMAL, &bg_color);

	vbox = gtk_vbox_new(FALSE, 10);

	announce_label = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(announce_label),
		"<big>Let's make a deal!</big>");
	gtk_box_pack_start(GTK_BOX(vbox), announce_label, FALSE, FALSE, 0);
	gtk_widget_show(announce_label);

	hbox = gtk_hbox_new(TRUE, 10);
	for (i = 0; i < NUM_DOORS; i++) {
		door = make_door_widget(i + 1);
		doors[i] = GTK_BUTTON(door);
		g_signal_connect(G_OBJECT(door), "clicked",
			G_CALLBACK(door_clicked), GUINT_TO_POINTER(i));
		gtk_box_pack_start(GTK_BOX(hbox), door, FALSE, FALSE, 0);
	}
	gtk_widget_show(hbox);
	box = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(box), hbox, TRUE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), box, TRUE, FALSE, 0);
	gtk_widget_show(box);

	hbox = gtk_hbox_new(FALSE, 10);
	summary_label = gtk_label_new(NULL);
	summary_label_update();
	gtk_box_pack_end(GTK_BOX(hbox), summary_label, FALSE, FALSE, 0);
	gtk_widget_show(summary_label);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	gtk_container_add(GTK_CONTAINER(window), vbox);
	gtk_widget_show(vbox);

	gtk_widget_show(window);

	gtk_main();

	/*
	for (;;) {
		game_print_report(&game);
		game_start_round(&game);
		game_choose_door(&game, 0);
		game_reveal_doors(&game);
		if (rand() % 2)
			game_switch(&game);
		game_finish_round(&game);
	}
	*/

	game_free(&game);

	return 0;
}
